#include "runtime/mem_arena.h"
#include "runtime/resources.h"
#include "runtime/threading.h"

#include "gfx/builtin_objects.h"
#include "gfx/gfx.h"
#include "gfx/renderer_api.h"
#include "gfx/effect.h"

#include "asset_compiler/asset_compiler.h"

void RegisterCVars(void) {
    rtRegisterAssetCompilerCVars();
}

static rt_render_graph *_graph;

static rt_result ForwardPassExecute(uint32_t pass_id,
                                    rt_command_buffer_handle cmdbuf,
                                    const rt_render_view *views,
                                    uint32_t view_count,
                                    void *userdata) {
    RT_ASSERT(view_count == 1, "Expected a single view for the main camera.");
    rt_render_view view = *views;
    RT_VERIFY(rtDoViewTypesMatchExact(view, &g_builtin_render_object_types.render_mesh, 1));

    rt_render_list meshes = view.lists[0];
    if (meshes.length == 0)
        return RT_SUCCESS;

    rt_pipeline_handle bound_pipeline =
        (RT_GET_RENDER_LIST_ELEMENT(meshes, rt_render_mesh, 0)).pipeline;
    g_renderer.CmdBindPipeline(cmdbuf, bound_pipeline);

    for (size_t mesh_idx = 0; mesh_idx < meshes.length; ++mesh_idx) {
        rt_render_mesh mesh = RT_GET_RENDER_LIST_ELEMENT(meshes, rt_render_mesh, mesh_idx);
        if (RT_COMPARE_RENDER_HANDLES(bound_pipeline, mesh.pipeline, !=)) {
            bound_pipeline = mesh.pipeline;
            g_renderer.CmdBindPipeline(cmdbuf, bound_pipeline);
        }
        g_renderer.CmdBindVertexBuffers(cmdbuf, 0, 1, &mesh.vbo, NULL);
        g_renderer.CmdDraw(cmdbuf, 0, mesh.vertex_count);
    }

    return RT_SUCCESS;
}

/* Called after the runtime has finished its initialization and before entering the main-loop*/
void Init(void) {
    rtLog("GAME", "Init");
    rtInitAssetCompiler();

    rtWaitForAssetProcessing();

    rtRegisterBuiltinRenderObjectTypes();

    rt_render_graph_builder builder = g_renderer.CreateRenderGraphBuilder();
    rt_attachment_info backbuffer   = {
          .name    = "backbuffer",
          .format  = RT_PIXEL_FORMAT_SWAPCHAIN,
          .width   = RT_RENDER_TARGET_SIZE_SWAPCHAIN,
          .height  = RT_RENDER_TARGET_SIZE_SWAPCHAIN,
          .samples = 1,
          .layers  = 1,
    };
    builder.AddRenderTarget(builder.obj, &backbuffer);

    rt_pass_info forward = {.name  = "forward",
                            .flags = RT_PASS_FLAG_EXECUTE_ALWAYS | RT_PASS_FLAG_GRAPHICS};
    builder.AddRenderPass(builder.obj, &forward);
    builder.AddColorOutput(builder.obj,
                           "forward",
                           "backbuffer",
                           RT_PASS_LOAD_MODE_CLEAR,
                           RT_PASS_WRITE_MODE_STORE,
                           (rt_color){.r = 1.f, .g = 0.f, .b = 1.f, .a = 1.f});
    builder.SetBackbuffer(builder.obj, "backbuffer");
    builder.BindRenderPass(builder.obj, "forward", ForwardPassExecute, NULL);
    if (builder.Build(builder.obj, &_graph) != RT_SUCCESS) {
        rtReportError("GAME", "Failed to build the render graph.");
        return;
    }

    g_renderer.DestroyRenderGraphBuilder(&builder);

    const rt_effect *effect;
    if (rtLoadEffect(rtGetResourceID("assets/shader/static_object.effect"), &effect) !=
        RT_SUCCESS) {
        rtReportError("GAME", "Oh noo...");
    }

    const rt_effect *effect2;
    if (rtLoadEffect(rtGetResourceID("assets/shader/static_object.effect"), &effect2) !=
        RT_SUCCESS) {
        rtReportError("GAME", "Oh noo...");
    }

    rtReleaseEffect(effect);
    rtReleaseEffect(effect2);
}

/* Called after exiting the main-loop and before the runtime starts its shutdown */
void Shutdown(void) {
    rtLog("GAME", "Shutdown");
    rtShutdownAssetCompiler();
}

// Question; How do we move data from update to render.
// This is where we could fill the render views, but that would
// mean double/triple buffering the views
void Update(unsigned int frame_id) {
    RT_UNUSED(frame_id);
}

void Render(unsigned int frame_id) {
    g_renderer.ResetRenderGraph(_graph);
    g_renderer.ExecuteRenderGraph(_graph);
}